#include "assets.h"
#include "asset_dependencies.h"
#include "aio.h"
#include "buffer_manager.h"
#include "config.h"
#include "threading.h"
#include "uidtab.h"

#include <assert.h>
#include <stdlib.h>

VY_CVAR_I(rt_AssetCacheSize, "Number of asset cache entries. Default: 1024.", 1024);

/* asset_loading.c */
extern vy_result DecompressAsset(void *compressed_buffer,
                                 size_t compressed_buffer_size,
                                 void **p_decompressed,
                                 size_t *p_decompressed_size);

typedef enum {
    CACHE_ENTRY_STATE_FREE,
    CACHE_ENTRY_STATE_LOADING,
    CACHE_ENTRY_STATE_LOADED,
} vy_asset_cache_entry_state;

typedef struct vy_asset_cache_entry_s {
    vy_asset_cache_entry_state state;
    vy_aio_handle load;
    void *buffer;
    size_t size;
    int refcount;

    /* Reclaim list */
    struct vy_asset_cache_entry_s *prev_reclaim;
    struct vy_asset_cache_entry_s *next_reclaim;
} vy_asset_cache_entry;

static vy_uid *_uids;
static vy_asset_cache_entry *_entries;
static vy_asset_cache_entry *_first_reclaim;
static vy_asset_cache_entry *_last_reclaim;

/* Locked as writer when modifiying entries, as reader when searching */
static vy_rwlock _lock;

vy_result InitAssetCache(void) {
    _entries = calloc((size_t)rt_AssetCacheSize.i, sizeof(vy_asset_cache_entry));
    if (!_entries) {
        return VY_BUFFER_ALLOC_FAILED;
    }
    _uids = calloc((size_t)rt_AssetCacheSize.i, sizeof(vy_uid));
    if (!_uids) {
        free(_entries);
        return VY_BUFFER_ALLOC_FAILED;
    }
    vy_create_rwlock_result lock_res = vyCreateRWLock();
    if (!lock_res.ok) {
        free(_entries);
        free(_uids);
        return VY_UNKNOWN_ERROR;
    }
    _lock = lock_res.lock;
    return VY_SUCCESS;
}

void ShutdownAssetCache(void) {
    free(_entries);
    free(_uids);
    vyDestroyRWLock(&_lock);
    _first_reclaim = NULL;
    _last_reclaim  = NULL;
}

static void ReleaseEntry(vy_asset_cache_entry *entry) {
    if (entry->load != VY_AIO_INVALID_HANDLE) {
        vyWaitForAIOCompletion(entry->load);
        vyReleaseAIO(entry->load);
        entry->load = VY_AIO_INVALID_HANDLE;
    }
    vyReleaseBuffer(entry->buffer, entry->size);
    entry->buffer       = NULL;
    entry->size         = 0;
    entry->next_reclaim = NULL;
    entry->prev_reclaim = NULL;
}

static void GarbageCollect(void) {
    vyLockWrite(&_lock);
    vy_asset_cache_entry *entry = _first_reclaim;
    while (entry) {
        assert(entry->refcount == 0);
        vy_asset_cache_entry *next = entry->next_reclaim;
        if (entry->state == CACHE_ENTRY_STATE_LOADED) {
            ReleaseEntry(entry);
            _first_reclaim = next;
        }
        entry          = next;
    }
    vyUnlockWrite(&_lock);
}

static vy_asset_cache_entry *GetEntry(vy_uid uid) {
    /* Hash lookup */
    unsigned int mod = (unsigned int)rt_AssetCacheSize.i - 1;
    for (unsigned int i = 0; i < (unsigned int)rt_AssetCacheSize.i; ++i) {
        unsigned int slot = (uid + i) & mod;
        if (_uids[slot] == uid) {
            return &_entries[slot];
        } else if (_uids[slot] == VY_INVALID_UID) {
            break;
        }
    }
    return NULL;
}


static bool IsAssetLoaded(vy_uid uid) {
    const vy_asset_cache_entry *entry = GetEntry(uid);
    if (entry)
        return entry->state == CACHE_ENTRY_STATE_LOADED ||
               entry->state == CACHE_ENTRY_STATE_LOADING;
    else
        return false;
}

static int InsertEntry(vy_uid uid) {
    unsigned int mod = (unsigned int)rt_AssetCacheSize.i - 1;
    for (unsigned int i = 0; i < (unsigned int)rt_AssetCacheSize.i; ++i) {
        unsigned int slot = (uid + i) & mod;
        if (_uids[slot] == 0) {
            return (int)slot;
        }
    }
    return -1;
}
static vy_result InsertAndLoadAssets(const vy_uid *uids, size_t count) {
    vy_load_batch batch = {.num_loads = 0};

    vy_result res = VY_SUCCESS;

    count = (count < VY_LOAD_BATCH_MAX_SIZE) ? count : VY_LOAD_BATCH_MAX_SIZE;
    vy_asset_cache_entry *load_entries[VY_LOAD_BATCH_MAX_SIZE];

    for (size_t i = 0; i < count; ++i) {
        vyLockRead(&_lock);
        bool needs_load = !IsAssetLoaded(uids[i]);
        vyUnlockRead(&_lock);

        if (!needs_load)
            continue;

        vyLockWrite(&_lock);
        /* It's possible that another thread loaded the asset in the meantime */
        if (!IsAssetLoaded(uids[i])) {
            const vy_uid_data *data = vyGetUIDData(uids[i]);
            if (!data) {
                vyUnlockWrite(&_lock);
                vyLog("ASSET_CACHE", "Failed to get uid data for uid %u", uids[i]);
                res = VY_UNKNOWN_ASSET;
                continue;
            }

            void *compressed_data = vyAllocBuffer(data->size);
            if (!compressed_data) {
                /* Try again after garbage collection */
                vyUnlockWrite(&_lock);
                GarbageCollect();
                compressed_data = vyAllocBuffer(data->size);
                if (!compressed_data) {
                    vyLog("ASSET_CACHE",
                          "Failed to allocate intermediate buffer for uid %u",
                          uids[i]);
                    res = VY_BUFFER_ALLOC_FAILED;
                    continue;
                }
                vyLockWrite(&_lock);
            }

            int slot = InsertEntry(uids[i]);
            if (slot == -1) {
                vyUnlockWrite(&_lock);
                vyLog("ASSET_CACHE", "Failed to insert new entry for uid %u", uids[i]);
                res = VY_ASSET_CACHE_FULL;
                break;
            }
            
            vy_asset_cache_entry *entry = &_entries[slot];
            load_entries[batch.num_loads] = entry;

            /* We set the refcount to 0, but don't insert the entry
             * into the reclaim list, to ensure that its buffer does not get freed
             * while the load still executes. Setting the refcount to 0 ensures
             * that the count is correct, once the asset is accessed the first time. */
            entry->state        = CACHE_ENTRY_STATE_LOADING;
            entry->refcount     = 0;
            entry->buffer       = compressed_data;
            entry->size         = data->size;
            entry->next_reclaim = NULL;
            entry->prev_reclaim = NULL;
            entry->load         = VY_AIO_INVALID_HANDLE;

            batch.loads[batch.num_loads].file      = data->pkg_file;
            batch.loads[batch.num_loads].num_bytes = data->size;
            batch.loads[batch.num_loads].offset    = data->offset;
            batch.loads[batch.num_loads].dest      = compressed_data;
            ++batch.num_loads;
        }
    }
    vyUnlockWrite(&_lock);

    /* Dispatch the load */
    vy_aio_handle handles[VY_LOAD_BATCH_MAX_SIZE];
    if ((res = vySubmitLoadBatch(&batch, handles)) != VY_SUCCESS) {
        vyLog("ASSET_CACHE", "Failed to submit %u asset loads.", batch.num_loads);
        return res;
    }

    /* Set the aio handles of the inserted entries */
    vyLockWrite(&_lock);
    for (unsigned int i = 0; i < batch.num_loads; ++i) {
        load_entries[batch.num_loads]->load = handles[i];
    }
    vyUnlockWrite(&_lock);

    return res;
}

static bool DecompressEntry(vy_uid uid, vy_asset_cache_entry *entry) {
    vyReleaseAIO(entry->load);
    entry->load = VY_AIO_INVALID_HANDLE;

    void *decompressed_buffer;
    size_t decompressed_size;
    vy_result dec_res =
        DecompressAsset(entry->buffer, entry->size, &decompressed_buffer, &decompressed_size);
    if (dec_res == VY_SUCCESS) {
        vyReleaseBuffer(entry->buffer, entry->size);
        entry->buffer = decompressed_buffer;
        entry->size   = decompressed_size;
        entry->state  = CACHE_ENTRY_STATE_LOADED;
        return true;
    } else if (dec_res == VY_BUFFER_ALLOC_FAILED) {
        GarbageCollect();
        /* Try again */
        if (DecompressAsset(entry->buffer, entry->size, &decompressed_buffer, &decompressed_size) ==
            VY_SUCCESS) {
            vyReleaseBuffer(entry->buffer, entry->size);
            entry->buffer = decompressed_buffer;
            entry->size   = decompressed_size;
            entry->state  = CACHE_ENTRY_STATE_LOADED;
            return true;
        }
        /* Don't do anything yet. We might be able to to do this later, once some
         * buffers become free. */
        vyLog("ASSET_CACHE", "Failed to decompress asset %u", uid);
        return false;
    } else {
        vyLog("ASSET_CACHE", "Failed to decompress asset %u", uid);
        ReleaseEntry(entry);

        ptrdiff_t idx = entry - _entries;
        _uids[idx]    = VY_INVALID_UID;
        return false;
    }
}

static void CheckCompletedLoads(const vy_uid *uids, size_t count) {
    for (size_t i = 0; i < count; ++i) {
        vyLockRead(&_lock);
        volatile vy_asset_cache_entry *entry = (volatile vy_asset_cache_entry *)GetEntry(uids[i]);
        if (!entry) {
            vyUnlockRead(&_lock);
            vyLog("ASSET_CACHE", "Passed unknown uid %u to CheckCompletedLoads()", uids[i]);
            continue;
        }

        if (entry->state != CACHE_ENTRY_STATE_LOADING) {
            vyUnlockRead(&_lock);
            continue;
        }
        bool load_finished = vyGetAIOState(entry->load) == VY_AIO_STATE_FINISHED;
        vyUnlockRead(&_lock);
       
        if (load_finished) {
            vyLockWrite(&_lock);
            /* Ensure that no-one else handled this */
            if (entry->state == CACHE_ENTRY_STATE_LOADING) {
                DecompressEntry(uids[i], (vy_asset_cache_entry *)entry);
            }
            vyUnlockWrite(&_lock);
        }
    }
}

VY_DLLEXPORT vy_get_asset_result vyGetAsset(vy_uid uid) {
    vy_get_asset_result result = {
        .result = VY_SUCCESS,
    };

    vyLockRead(&_lock);
    bool needs_load = !IsAssetLoaded(uid);
    vyUnlockRead(&_lock);

    if (needs_load) {
        vy_uid load_uids[VY_LOAD_BATCH_MAX_SIZE];
        size_t load_count = 1;
        load_uids[0]      = uid;

        vy_asset_dependency_list deps = vyGetAssetDependencies(uid);
        for (size_t i = 0; i < deps.count && i < VY_LOAD_BATCH_MAX_SIZE - 1; ++i) {
            load_uids[i + 1] = deps.dependencies[i];
            ++load_count;
        }

        result.result = InsertAndLoadAssets(load_uids, load_count);
        if (result.result == VY_SUCCESS) {
            CheckCompletedLoads(load_uids, load_count);
        }
    }

    vyLockRead(&_lock);
    volatile vy_asset_cache_entry *entry = GetEntry(uid);
    if (entry) {
        if (entry->state == CACHE_ENTRY_STATE_LOADED) {
            ++entry->refcount;
            result.data = entry->buffer;
            result.size = entry->size;
        } else if (entry->state == CACHE_ENTRY_STATE_LOADING) {
            /* Promote to write lock */
            vyUnlockRead(&_lock);
            vyLockWrite(&_lock);
            if (entry->state == CACHE_ENTRY_STATE_LOADING) {
                assert(entry->load != VY_AIO_INVALID_HANDLE);
                ++entry->refcount;
                if (vyWaitForAIOCompletion(entry->load) == VY_AIO_STATE_FINISHED) {
                    if (DecompressEntry(uid, (vy_asset_cache_entry *)entry)) {
                        result.data = entry->buffer;
                        result.size = entry->size;
                    } else {
                        result.result = VY_LOAD_FAILED;
                    }
                } else {
                    ReleaseEntry((vy_asset_cache_entry *)entry);
                    vyLog("ASSET_CACHE", "Failed to load asset %u", uid);
                    result.result = VY_LOAD_FAILED;
                }
            }
            vyUnlockWrite(&_lock);

            /* To match the unlock below */
            vyLockRead(&_lock);
        }
    }
    vyUnlockRead(&_lock);

    return result;
}