#include "assets.h"
#include "uidtab.h"
#include "aio.h"
#include "buffer_manager.h"

#define VY_DEFINE_PACKAGE_FILE_STRUCTURES
#include "packages.h"

#include "lz4/lz4.h"

vy_result DecompressAsset(void *compressed_buffer,
                          size_t compressed_buffer_size,
                          void **p_decompressed,
                          size_t *p_decompressed_size) {

    const vy_package_asset_header *header = compressed_buffer;

    size_t compressed_size       = (compressed_buffer_size) - sizeof(*header);
    XXH64_hash_t calculated_hash = XXH3_64bits((header + 1), compressed_size);
    XXH64_hash_t file_hash       = XXH64_hashFromCanonical(&header->checksum);
    if (calculated_hash != file_hash) {
        vyLog("core", "Checksum mismatch for asset");
        return VY_LOAD_FAILED;
    }

    size_t size               = (size_t)header->decompressed_size;
    void *decompressed_buffer = vyAllocBuffer(size);
    if (!decompressed_buffer) {
        return VY_BUFFER_ALLOC_FAILED;
    }

    if (LZ4_decompress_safe((const char *)(header + 1),
                            (char *)decompressed_buffer,
                            (int)compressed_size,
                            (int)size) < 0) {
        return VY_UNKNOWN_ERROR;
    }

    *p_decompressed      = decompressed_buffer;
    *p_decompressed_size = size;
    return VY_SUCCESS;
}


VY_DLLEXPORT vy_result vyLoadAssetDirect(vy_uid uid, void **p_buffer, size_t *p_size) {
    const vy_uid_data *data = vyGetUIDData(uid);
    if (!data)
        return VY_UNKNOWN_ASSET;

    void *compressed_buffer = vyAllocBuffer(data->size);
    if (!compressed_buffer) {
        return VY_BUFFER_ALLOC_FAILED;
    }

    if (vySubmitSingleLoadSync((vy_file_load) {
            .file      = data->pkg_file,
            .offset    = data->offset,
            .num_bytes = data->size,
            .dest      = compressed_buffer,
        }) != VY_AIO_STATE_FINISHED) {
        vyReleaseBuffer(compressed_buffer, data->size);
        return VY_LOAD_FAILED;
    }
    
    void *decompressed_buffer;
    size_t decompressed_size;
    vy_result res = DecompressAsset(compressed_buffer, data->size, &decompressed_buffer, &decompressed_size);

    vyReleaseBuffer(compressed_buffer, data->size);
    *p_buffer = decompressed_buffer;
    *p_size   = decompressed_size;

    return res;
}