#include "assetmeta.h"
#include "processing.h"
#include "utils.h"
#include "packages.h"
#include "dependency_tracking.h"

#define VY_ASSETC_DONT_DEFINE_OPTIONS_GLOBAL
#include "options.h"

#include "runtime/aio.h"
#include "runtime/buffer_manager.h"
#include "runtime/uidtab.h"

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

extern vy_result vyProcessPipelineFile(vy_file_id file,
                                       void *buffer,
                                       size_t size,
                                       uint32_t flags,
                                       vy_processor_output *output);

extern vy_result vyProcessShaderFile(vy_file_id file,
                                     void *buffer,
                                     size_t size,
                                     uint32_t flags,
                                     vy_processor_output *output);

extern void vyDiscoverAssets(void);

vy_assetc_options g_assetc_options = {
    .root_directory   = ".",
    .optimization     = VY_ASSET_OPTIMIZATION_NONE,
    .renderer_backend = VY_RENDERER_BACKEND_CODE_VK,
};

static bool ParseCommandLineArgs(int argc, char **argv) {
    bool have_root_directory = false;
    for (int i = 1; i < argc; ++i) {
        if (i < argc - 1) {
            const char *val = argv[i + 1];
            bool matched    = false;

            if (strcmp(argv[i], "-r") == 0 || strcmp(argv[i], "--renderer") == 0) {
                if (strcmp(val, "vk") == 0) {
                    g_assetc_options.renderer_backend = VY_RENDERER_BACKEND_CODE_VK;
                } else {
                    vyReportError("ASSETC",
                                  "Invalid render backend %s. Valid options are: vk",
                                  val);
                    return false;
                }
                matched = true;
            } else if (strcmp(argv[i], "-o") == 0 || strcmp(argv[i], "--optimization") == 0) {
                if (strcmp(val, "none") == 0) {
                    g_assetc_options.optimization = VY_ASSET_OPTIMIZATION_NONE;
                } else if (strcmp(val, "space") == 0) {
                    g_assetc_options.optimization = VY_ASSET_OPTIMIZATION_SPACE;
                } else if (strcmp(val, "performance") == 0) {
                    g_assetc_options.optimization = VY_ASSET_OPTIMIZATION_PERFORMANCE;
                } else {
                    vyReportError("ASSETC",
                                  "Invalid optimization level %s. Valid options are: none, space, "
                                  "performance",
                                  val);
                    return false;
                }
                matched = true;
            } else if (argv[i][0] == '-') {
                vyReportError("ASSETC", "Invalid command line argument %s", argv[i]);
                return false;
            }

            if (matched) {
                /* Skip the value */
                ++i;
                continue;
            }
        }

        if (!have_root_directory) {
            g_assetc_options.root_directory = argv[i];
        } else {
            /* Maybe have secondary directories later? */
            vyReportError("ASSETC",
                          "More than one root directory passed as command line "
                          "argument.");
            return false;
        }
    }
    return true;
}

int main(int argc, char **argv) {
    vyInitRuntime();

    /* Init assetc */
    if (!ParseCommandLineArgs(argc, argv)) {
        return 1;
    }
    vyInitPackages();
    

    if (vyLoadAssetMeta() != VY_SUCCESS) {
        return 1;
    }
    
    if (vyInitDependencyTracking() != VY_SUCCESS) {
        return 1;
    }

    if (vyAddAssetProcessor(".pipeline", vyProcessPipelineFile) != VY_SUCCESS)
        return 1;
    if (vyAddAssetProcessor(".glsl", vyProcessShaderFile) != VY_SUCCESS)
        return 1;

    if (vyStartProcessing() != VY_SUCCESS) {
        return 1;
    }

    /* Create necessary directories */
    vySetWorkingDirectory(g_assetc_options.root_directory);
    vyCreateDirectory("assets");
    vyCreateDirectory("actemp");
    vyCreateDirectory("data");

    /* "Mainloop" */
    vyDiscoverAssets();

    vyWaitUntilProcessingIsFinished();
    vyStopProcessing();

    /* Write result */
    vySaveAssetMeta();
    if (vySavePackages() != VY_SUCCESS) {
        return 1;
    }
    if (vyWriteUIDTab() != VY_SUCCESS) {
        return 1;
    }
    if (vySaveAssetDependencies() != VY_SUCCESS) {
        return 1;
    }

    vyShutdownRuntime();

    return 0;
}