#include "description_parser.h"

#include "runtime/runtime.h"

#include <stdbool.h>
#include <limits.h>
#include <stdlib.h>

static bool IsAllowedChar(char c) {
    return (c >= '0' && c <= '9') || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') ||
           (c == '.') || (c == '_') || (c == '/');
}

static bool IsWhitespace(char c) {
    return c == ' ' || c == '\t' || c == '\n' || c == '\r';
}

static void SkipWhitespace(vy_parse_state *state) {
    while (state->at < state->length && IsWhitespace(state->text[state->at])) {
        if (state->text[state->at] == '\n')
            ++state->line;
        ++state->at;
    }
}

static bool ParseAttribute(vy_parse_state *state, vy_text_span *_name) {
    vy_text_span name;
    name.start  = &state->text[state->at];
    name.length = 0;
    while (state->at < state->length && !IsWhitespace(state->text[state->at])) {
        if (IsAllowedChar(state->text[state->at])) {
            ++state->at;
            ++name.length;
        } else {
            vyReportError("GFX",
                          "%s:%d Unexpected character %c",
                          state->file,
                          state->line,
                          state->text[state->at]);
            return false;
        }
    }
    *_name = name;
    return true;
}

static bool ParseValue(vy_parse_state *state, vy_text_span *_value) {
    vy_text_span value;
    value.start  = &state->text[state->at];
    value.length = 0;
    while (state->at < state->length && !IsWhitespace(state->text[state->at]) &&
           state->text[state->at] != ';') {
        if (IsAllowedChar(state->text[state->at])) {
            ++state->at;
            ++value.length;
        } else {
            vyReportError("GFX",
                          "%s:%d Unexpected character %c",
                          state->file,
                          state->line,
                          state->text[state->at]);
            return false;
        }
    }
    *_value = value;
    return true;
}

static bool ParseStmtList(vy_parse_state *state, unsigned int *list_index);

static bool ParseStmt(vy_parse_state *state, unsigned int *stmt_index) {
    vy_parsed_stmt stmt;
    stmt.next = UINT_MAX; /* end of list */

    SkipWhitespace(state);
    if (!ParseAttribute(state, &stmt.attribute))
        return false;
    SkipWhitespace(state);

    if (state->at == state->length) {
        vyReportError("GFX", "%s:%d Expected either a value or '{'", state->file, state->line);
        return false;
    }

    if (state->text[state->at] == '{') {
        /* Consume '{' */
        ++state->at;

        stmt.form = VY_STMT_FORM_LIST;
        if (!ParseStmtList(state, &stmt.list_index))
            return false;

        /* Consume '}' */
        if (state->at < state->length && state->text[state->at] == '}')
            ++state->at;
    } else {
        stmt.form = VY_STMT_FORM_VALUE;
        if (!ParseValue(state, &stmt.value))
            return false;

        /* Consume ';' */
        if (state->at < state->length && state->text[state->at] == ';')
            ++state->at;
    }

    SkipWhitespace(state);

    /* Add statement to array */
    if (state->statement_count == state->statement_capacity) {
        unsigned int cap     = (state->statement_capacity > 0) ? state->statement_capacity * 2 : 64;
        vy_parsed_stmt *temp = realloc(state->statements, sizeof(vy_parsed_stmt) * cap);
        if (!temp) {
            vyReportError("GFX", "While parsing %s: Out of memory\n", state->file);
            return false;
        }
        state->statements         = temp;
        state->statement_capacity = cap;
    }
    state->statements[state->statement_count] = stmt;
    *stmt_index                               = state->statement_count++;

    return true;
}

static bool ParseStmtList(vy_parse_state *state, unsigned int *list_index) {
    vy_parsed_stmt_list list;
    list.first = UINT_MAX;
    list.count = 0;

    unsigned int last = UINT_MAX;

    while (state->at < state->length && state->text[state->at] != '}') {
        unsigned int stmt;
        if (!ParseStmt(state, &stmt))
            return false;

        if (last != UINT_MAX)
            state->statements[last].next = stmt;
        else
            list.first = stmt;
        last = stmt;

        ++list.count;
    }

    /* Add list to array */
    if (state->statement_list_count == state->statement_list_capacity) {
        unsigned int cap =
            (state->statement_list_capacity > 0) ? state->statement_list_capacity * 2 : 64;
        vy_parsed_stmt_list *temp =
            realloc(state->statement_lists, sizeof(vy_parsed_stmt_list) * cap);
        if (!temp) {
            vyReportError("GFX", "While parsing %s: Out of memory\n", state->file);
            return false;
        }
        state->statement_lists         = temp;
        state->statement_list_capacity = cap;
    }
    state->statement_lists[state->statement_list_count] = list;
    *list_index                                         = state->statement_list_count++;
    return true;
}

const vy_parsed_stmt * vyFindStatement(const vy_parse_state *state, unsigned int list_index, const char *attribute) {
    if (list_index >= state->statement_list_count)
        return NULL;
    const vy_parsed_stmt_list *list = &state->statement_lists[list_index];
    unsigned int stmt_index         = list->first;
    for (unsigned int i = 0; i < list->count; ++i) {
        const vy_parsed_stmt *stmt = &state->statements[stmt_index];
        if (vyCompareSpanToString(stmt->attribute, attribute) == 0)
            return stmt;
        stmt_index = stmt->next;
    }
    return NULL;
}


vy_result vyParseDescription(const char *text,
                             size_t length,
                             const char *file_path,
                             unsigned int *_root_list,
                             vy_parse_state *_state) {

    vy_parse_state state = {.text                    = text,
                            .at                      = 0,
                            .length                  = length,
                            .line                    = 1,
                            .file                    = file_path,
                            .statements              = NULL,
                            .statement_lists         = NULL,
                            .statement_capacity      = 0,
                            .statement_list_capacity = 0};

    unsigned int root_list = 0;
    if (!ParseStmtList(&state, &root_list)) {
        return 1;
    }
    *_root_list = root_list;
    *_state     = state;

    return VY_SUCCESS;
}


void vyReleaseParseState(vy_parse_state *state) {
    free(state->statements);
    free(state->statement_lists);
}