#include "packages.h"
#include "processing.h"
#include "utils.h"

#define VY_DEFINE_PACKAGE_FILE_STRUCTURES
#include "runtime/threading.h"
#include "runtime/assets.h"
#include "runtime/file_tab.h"
#include "runtime/packages.h"

#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdbool.h>

#include "xxhash/xxhash.h"
#include "lz4/lz4.h"

typedef struct {
    vy_uid uid;
    size_t disk_size;
} vy_package_entry;

typedef struct {
    char *name;
    unsigned int num_entries;
    unsigned int entry_capacity;
    vy_package_entry *entries;
} vy_package;

#define MAX_PACKAGES 1024

vy_package _packages[MAX_PACKAGES];
unsigned int _package_count = 0;

vy_mutex *_guard;

unsigned int vyAddPackageFile(vy_text_span name) {
    vyLockMutex(_guard);

    for (unsigned int i = 0; i < _package_count; ++i) {
        if (vyCompareSpanToString(name, _packages[i].name + 5) == 0) {
            vyUnlockMutex(_guard);
            return i;
        }
    }

    /* Create a new package */
    _packages[_package_count].name = malloc(name.length + 1 + 5);
    if (!_packages[_package_count].name) {
        vyReportError("ASSETC",
                      "Failed to allocate storage for new package %*.s",
                      name.length,
                      name.start);
        vyUnlockMutex(_guard);
        return UINT_MAX;
    }
    memcpy(_packages[_package_count].name, "data/", 5);
    memcpy(_packages[_package_count].name + 5, name.start, name.length);
    _packages[_package_count].name[name.length + 5] = '\0';

    unsigned int index = _package_count++;

    vyUnlockMutex(_guard);
    return index;
}

void vyInitPackages(void) {
    _guard = vyCreateMutex();
    /* Create the default package (0) */
    vyAddPackageFile((vy_text_span){.start = "default.pkg", .length = 12});
}

static void AddAssetToPackageImpl(unsigned int package,
                                  vy_uid uid,
                                  void *buffer,
                                  size_t size,
                                  bool needs_compression) {
    vyLockMutex(_guard);
    if (package >= _package_count) {
        vyReportError("ASSETC", "Trying to add an asset to a non-existing package.");
        vyUnlockMutex(_guard);
        return;
    }

    vy_package *pkg = &_packages[package];

    for (unsigned int i = 0; i < pkg->num_entries; ++i) {
        if (pkg->entries[i].uid == uid) {
            vyUnlockMutex(_guard);
            return;
        }
    }

    if (pkg->num_entries == pkg->entry_capacity) {
        unsigned int new_cap = (pkg->entry_capacity > 0) ? 2 * pkg->entry_capacity : 256;
        vy_package_entry *n  = realloc(pkg->entries, new_cap * sizeof(vy_package_entry));
        if (!n) {
            vyReportError("ASSETC", "Failed to grow storage for package %u.", package);
            return;
        }
        pkg->entry_capacity = new_cap;
        pkg->entries        = n;
    }

    char tmp_path[256];
    snprintf(tmp_path, 256, "actemp/%u.bin", uid);
    if (needs_compression) {
        FILE *tmp_f = fopen(tmp_path, "wb");
        if (!tmp_f) {
            vyUnlockMutex(_guard);
            return;
        }

        int required_size       = LZ4_compressBound((int)size);
        void *compressed_buffer = malloc(required_size);
        if (!compressed_buffer) {
            fclose(tmp_f);
            vyUnlockMutex(_guard);
            return;
        }

        int compressed_bytes =
            LZ4_compress_default(buffer, compressed_buffer, (int)size, required_size);
        if (compressed_bytes == 0) {
            free(compressed_buffer);
            fclose(tmp_f);
            vyReportError("ASSETC", "Failed to compress asset %x of package %s", uid, pkg->name);
            return;
        }

        vy_package_asset_header header;
        XXH64_hash_t checksum = XXH3_64bits_withSeed(buffer, (size_t)compressed_bytes, 0);
        XXH64_canonicalFromHash(&header.checksum, checksum);
        header.decompressed_size = (uint32_t)size;

        if (fwrite(&header, sizeof(header), 1, tmp_f) != 1) {
            vyReportError("ASSETC", "Failed to write to actemp/%u.bin", uid);
            free(compressed_buffer);
            fclose(tmp_f);
            return;
        }
        if (fwrite(buffer, compressed_bytes, 1, tmp_f) != 1) {
            vyReportError("ASSETC", "Failed to write to actemp/%u.bin", uid);
            free(compressed_buffer);
            fclose(tmp_f);
            return;
        }
        fclose(tmp_f);

        pkg->entries[pkg->num_entries].disk_size =
            (size_t)compressed_bytes + sizeof(vy_package_asset_header);
    } else {
        pkg->entries[pkg->num_entries].disk_size = vyGetFileSize(tmp_path);
        if (pkg->entries[pkg->num_entries].disk_size == 0) {
            vyReportError("ASSETC", "Failed to determine size of actemp/%u.bin", uid);
        }
    }
    
    pkg->entries[pkg->num_entries].uid               = uid;
    ++pkg->num_entries;


    vyUnlockMutex(_guard);
}

void vyAddAssetToPackage(unsigned int package, vy_uid uid, void *buffer, size_t size) {
    AddAssetToPackageImpl(package, uid, buffer, size, true);
}

void vyAddUnchangedAssetToPackage(unsigned int package, vy_uid uid) {
    AddAssetToPackageImpl(package, uid, NULL, 0, false);
}

static vy_result SavePackage(vy_package *pkg) {
    if (pkg->num_entries == 0) {
        vyLog("ASSETC", "Package %s has no entries.", pkg->name);
        return VY_SUCCESS;
    }

    size_t current_buffer_size = 0;
    void *buffer               = NULL;

    size_t offset_in_file = 0;

    vy_file_id package_fid = vyAddFile(pkg->name);

    FILE *f = fopen(pkg->name, "wb");
    if (!f) {
        vyReportError("ASSETC", "Failed to open %s for writing.", pkg->name);
        return VY_UNKNOWN_ERROR;
    }

    for (unsigned int i = 0; i < pkg->num_entries; ++i) {
        char tmp_path[256];
        snprintf(tmp_path, 256, "actemp/%u.bin", pkg->entries[i].uid);
        FILE *tmp_f = fopen(tmp_path, "rb");
        if (!tmp_f) {
            vyReportError("ASSETC", "Failed to open %s for reading.", tmp_path);
            fclose(f);
            free(buffer);
            return VY_UNKNOWN_ERROR;
        }

        if (current_buffer_size < pkg->entries[i].disk_size) {
            void *tmp = realloc(buffer, pkg->entries[i].disk_size);
            if (!tmp) {
                vyReportError("ASSETC", "Failed to allocate buffer (%zu bytes) for reading %s.", pkg->entries[i].disk_size, tmp_path);
                fclose(f);
                fclose(tmp_f);
                free(buffer);
                return VY_UNKNOWN_ERROR;
            }
            buffer = tmp;
            current_buffer_size = pkg->entries[i].disk_size;
        }

        if (fread(buffer, pkg->entries[i].disk_size, 1, tmp_f) != 1) {
            vyReportError("ASSETC", "Failed to read %s.", tmp_path);
            fclose(f);
            fclose(tmp_f);
            free(buffer);
            return VY_UNKNOWN_ERROR;
        }

        if (fwrite(buffer, pkg->entries[i].disk_size, 1, f) != 1) {
            vyReportError("ASSETC", "Failed to write (%zu bytes) to %s.", pkg->entries[i].disk_size, pkg->name);
            fclose(f);
            fclose(tmp_f);
            free(buffer);
            return VY_UNKNOWN_ERROR;
        }

        vyAddUIDTabEntry(package_fid,
                         pkg->entries[i].uid,
                         offset_in_file,
                         pkg->entries[i].disk_size);
        offset_in_file += pkg->entries[i].disk_size;

        fclose(tmp_f);
    }

    free(buffer);
    fclose(f);

    return VY_SUCCESS;
}

vy_result vySavePackages(void) {
    assert(vyIsMainThread());

    /* Save a .txt file with one line per package.
     * Enables us to re-init the file-tab in future runs. */
    FILE *f = fopen("data/packages.txt", "w");
    if (!f) {
        vyReportError("ASSETC", "Failed to write to 'packages.txt'");
        return VY_UNKNOWN_ERROR;
    }
    for (unsigned int i = 0; i < _package_count; ++i) {
        if (_packages[i].num_entries == 0)
            continue;
        fprintf(f, "%s\n", _packages[i].name);
    }
    fclose(f);

    for (unsigned int i = 0; i < _package_count; ++i) {
        vy_result res = SavePackage(&_packages[i]);
        if (res != VY_SUCCESS)
            return res;
    }
    return VY_SUCCESS;
}